@@ -7,7 +7,6 @@ local glob       = require 'glob'
 local platform   = require 'bee.platform'
 local await      = require 'await'
 local client     = require 'client'
-local plugin     = require 'plugin'
 local util       = require 'utility'
 local fw         = require 'filewatch'
 local scope      = require 'workspace.scope'
@@ -47,15 +46,29 @@ end
 --- 初始化工作区
 function m.create(uri)
     log.info('Workspace create: ', uri)
-    if uri == furi.encode '/'
-    or uri == furi.encode(os.getenv 'HOME') then
-        client.showMessage('Error', lang.script('WORKSPACE_NOT_ALLOWED', furi.decode(uri)))
-        return
-    end
-    local path = m.normalize(furi.decode(uri))
-    fw.watch(path)
     local scp = scope.createFolder(uri)
     m.folders[#m.folders+1] = scp
+    if uri == furi.encode '/'
+    or uri == furi.encode(os.getenv 'HOME' or '') then
+        client.showMessage('Error', lang.script('WORKSPACE_NOT_ALLOWED', furi.decode(uri)))
+        scp:set('bad root', true)
+    end
+end
+
+function m.remove(uri)
+    log.info('Workspace remove: ', uri)
+    for i, scp in ipairs(m.folders) do
+        if scp.uri == uri then
+            scp:remove()
+            table.remove(m.folders, i)
+            scp:set('ready', false)
+            scp:set('nativeMatcher', nil)
+            scp:set('libraryMatcher', nil)
+            scp:removeAllLinks()
+            m.flushFiles(scp)
+            return
+        end
+    end
 end
 
 function m.reset()
@@ -74,10 +87,9 @@ local globInteferFace = {
     type = function (path)
         local result
         pcall(function ()
-            local status = fs.symlink_status(path):type()
-            if status == 'directory' then
+            if fs.is_directory(path) then
                 result = 'directory'
-            elseif status == 'regular' then
+            else
                 result = 'file'
             end
         end)
@@ -85,7 +97,7 @@ local globInteferFace = {
     end,
     list = function (path)
         local fullPath = fs.path(path)
-        if fs.symlink_status(fullPath):type() ~= 'directory' then
+        if not fs.is_directory(fullPath) then
             return nil
         end
         local paths = {}
@@ -154,7 +166,7 @@ function m.getNativeMatcher(scp)
     end
 
     local matcher = glob.gitignore(pattern, {
-        root       = furi.decode(scp.uri),
+        root       = scp.uri and furi.decode(scp.uri),
         ignoreCase = platform.OS == 'Windows',
     }, globInteferFace)
 
@@ -189,9 +201,12 @@ function m.getLibraryMatchers(scp)
             librarys[m.normalize(path)] = true
         end
     end
-    log.debug('meta path:', scp:get 'metaPath')
-    if scp:get 'metaPath' then
-        librarys[m.normalize(scp:get 'metaPath')] = true
+    local metaPaths = scp:get 'metaPaths'
+    log.debug('meta path:', inspect(metaPaths))
+    if metaPaths then
+        for _, metaPath in ipairs(metaPaths) do
+            librarys[m.normalize(metaPath)] = true
+        end
     end
 
     local matchers = {}
@@ -216,7 +231,6 @@ function m.getLibraryMatchers(scp)
 end
 
 --- 文件是否被忽略
----@async
 ---@param uri uri
 function m.isIgnored(uri)
     local scp    = scope.getScope(uri)
@@ -253,7 +267,7 @@ function m.awaitLoadFile(uri)
         scp:get('cachedUris')[uri] = true
         ld:loadFile(uri)
     end)
-    ld:loadAll()
+    ld:loadAll(uri)
 end
 
 function m.removeFile(uri)
@@ -279,6 +293,10 @@ function m.awaitPreload(scp)
 
     scp:flushGC()
 
+    if scp:isRemoved() then
+        return
+    end
+
     local ld <close> = loading.create(scp)
     scp:set('loading', ld)
 
@@ -287,15 +305,22 @@ function m.awaitPreload(scp)
     local native   = m.getNativeMatcher(scp)
     local librarys = m.getLibraryMatchers(scp)
 
-    if scp.uri then
+    if scp.uri and not scp:get('bad root') then
         log.info('Scan files at:', scp:getName())
+        scp:gc(fw.watch(m.normalize(furi.decode(scp.uri)), true, function (path)
+            local uri = furi.encode(path)
+            if m.isIgnored(uri) and not files.isLibrary(uri) then
+                return false
+            end
+            return true
+        end))
         local count = 0
         ---@async
         native:scan(furi.decode(scp.uri), function (path)
             local uri = files.getRealUri(furi.encode(path))
             scp:get('cachedUris')[uri] = true
             ld:loadFile(uri)
-        end, function () ---@async
+        end, function (_) ---@async
             count = count + 1
             if count == 100000 then
                 client.showMessage('Warning', lang.script('WORKSPACE_SCAN_TOO_MUCH', count, furi.decode(scp.uri)))
@@ -306,6 +331,13 @@ function m.awaitPreload(scp)
     for _, libMatcher in ipairs(librarys) do
         log.info('Scan library at:', libMatcher.uri)
         local count = 0
+        scp:gc(fw.watch(furi.decode(libMatcher.uri), true, function (path)
+            local uri = furi.encode(path)
+            if m.isIgnored(uri) and not files.isLibrary(uri) then
+                return false
+            end
+            return true
+        end))
         scp:addLink(libMatcher.uri)
         ---@async
         libMatcher.matcher:scan(furi.decode(libMatcher.uri), function (path)
@@ -318,14 +350,13 @@ function m.awaitPreload(scp)
                 client.showMessage('Warning', lang.script('WORKSPACE_SCAN_TOO_MUCH', count, furi.decode(libMatcher.uri)))
             end
         end)
-        scp:gc(fw.watch(furi.decode(libMatcher.uri)))
     end
 
     -- must wait for other scopes to add library
     await.sleep(0.1)
 
     log.info(('Found %d files at:'):format(ld.max), scp:getName())
-    ld:loadAll()
+    ld:loadAll(scp:getName())
     log.info('Preload finish at:', scp:getName())
 end
 
@@ -354,9 +385,6 @@ end
 ---@param path string
 ---@return string
 function m.normalize(path)
-    if not path then
-        return nil
-    end
     path = path:gsub('%$%{(.-)%}', function (key)
         if key == '3rd' then
             return (ROOT / 'meta' / '3rd'):string()
@@ -368,6 +396,16 @@ function m.normalize(path)
     end)
     path = util.expandPath(path)
     path = path:gsub('^%.[/\\]+', '')
+    for _ = 1, 1000 do
+        if path:sub(1, 2) == '..' then
+            break
+        end
+        local count
+        path, count = path:gsub('[^/\\]+[/\\]+%.%.[/\\]', '/', 1)
+        if count == 0 then
+            break
+        end
+    end
     if platform.OS == 'Windows' then
         path = path:gsub('[/\\]+', '\\')
                    :gsub('[/\\]+$', '')
@@ -379,11 +417,10 @@ function m.normalize(path)
     return path
 end
 
----@return string
+---@param folderUri? uri
+---@param path string
+---@return string?
 function m.getAbsolutePath(folderUri, path)
-    if not path or path == '' then
-        return nil
-    end
     path = m.normalize(path)
     if fs.path(path):is_relative() then
         if not folderUri then
@@ -397,6 +434,7 @@ end
 
 ---@param uriOrPath uri|string
 ---@return string
+---@return boolean suc
 function m.getRelativePath(uriOrPath)
     local path, uri
     if uriOrPath:sub(1, 5) == 'file:' then
@@ -446,16 +484,24 @@ function m.flushFiles(scp)
     for uri in pairs(cachedUris) do
         files.delRef(uri)
     end
+    collectgarbage()
+    collectgarbage()
+    -- TODO: wait maillist
+    collectgarbage 'restart'
 end
 
 ---@param scp scope
 function m.resetFiles(scp)
     local cachedUris = scp:get 'cachedUris'
-    if not cachedUris then
-        return
+    if cachedUris then
+        for uri in pairs(cachedUris) do
+            files.resetText(uri)
+        end
     end
-    for uri in pairs(cachedUris) do
-        files.resetText(uri)
+    for uri in pairs(files.openMap) do
+        if scope.getScope(uri) == scp then
+            files.resetText(uri)
+        end
     end
 end
 
@@ -467,7 +513,7 @@ function m.awaitReload(scp)
     scp:set('libraryMatcher', nil)
     scp:removeAllLinks()
     m.flushFiles(scp)
-    plugin.init(scp)
+    m.onWatch('startReload', scp.uri)
     m.awaitPreload(scp)
     scp:set('ready', true)
     local waiting = scp:get('waitingReady')
@@ -501,11 +547,26 @@ function m.awaitReady(uri)
 end
 
 ---@param uri uri
+---@return boolean
 function m.isReady(uri)
     local scp = scope.getScope(uri)
     return scp:get('ready') == true
 end
 
+---@return boolean
+function m.isAllReady()
+    local scp = scope.fallback
+    if not scp:get 'ready' then
+        return false
+    end
+    for _, folder in ipairs(scope.folders) do
+        if not folder:get 'ready' then
+            return false
+        end
+    end
+    return true
+end
+
 function m.getLoadingProcess(uri)
     local scp = scope.getScope(uri)
     ---@type workspace.loading
@@ -520,9 +581,12 @@ end
 config.watch(function (uri, key, value, oldValue)
     if key:find '^Lua.runtime'
     or key:find '^Lua.workspace'
+    or key:find '^Lua.type'
     or key:find '^files' then
         if value ~= oldValue then
-            m.reload(scope.getScope(uri))
+            local scp = scope.getScope(uri)
+            m.reload(scp)
+            m.resetFiles(scp)
         end
     end
 end)
